from PyQt5.QtWidgets import QGraphicsItem

from effigy.NodeIO import NodeIO

class QNodeSceneNode(QGraphicsItem):
    """Graphical representation of Node"""

    # Here we define some information about the node
    author = "DrLuke"       # Author of this node (only used for namespacing, never visible to users)
    name = "Basenode"       # Human-readable name

    placeable = False       # Whether or not this node should be placeable from within the editor
    Category = ["Builtin"]     # Nested categories this Node should be sorted in

    # Description, similar to python docstrings (Brief summary on first line followed by a long description)
    description = """This node is the base class for all nodes.
It should never be placeable in the editor. However if you DO see this in the editor, something went wrong!"""

    def __init__(self, deserializedata=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.IO = []
        """Stores all IO for this Node."""

        self.setFlag(QGraphicsItem.ItemIsMovable)   # Item can be dragged with left-click

        if deserializedata is not None:
            self.deserialize(deserializedata)
        else:
            self.addGraphicsItems()
            self.addIO()

    def deserialize(self, data):
        """Gets data that has been serialized with 'serialize', reconstructs the object state from that data"""
        raise NotImplementedError("This method must be implemented in derived class")

    def addIO(self):
        """Add IOs to Node"""
        raise NotImplementedError("This method must be implemented in derived class")

    def addGraphicsItems(self):
        """Add Graphic Items for rendering the node"""
        raise NotImplementedError("This method must be implemented in derived class")

    def boundingRect(self):
        """Define appropriate bounding rect for Node body (necessary for interaction like dragging!)"""
        raise NotImplementedError("This method must be implemented in derived class")

    def paint(self, *__args):
        pass

    def mouseMoveEvent(self, QGraphicsSceneMouseEvent):
        for IO in self.IO:
            for link in IO.nodeLinks:
                link.updateBezier()

        super().mouseMoveEvent(QGraphicsSceneMouseEvent)





from PyQt5.QtWidgets import QGraphicsItem

from PyQt5.QtWidgets import QGraphicsRectItem, QGraphicsPathItem
from PyQt5.QtCore import QRectF, QPointF
from PyQt5.QtGui import QBrush, QColor, QPainterPath, QPen

from enum import Enum

class NodeIODirection(Enum):
    """Establish if IO direction is input, output or any"""
    any = 0
    output = 1
    input = 2

class NodeIOMultiplicity(Enum):
    """Whether the node can be connected to multiple other nodes, or just one"""
    single = 0
    multiple = 1

class InvalidLinkException(Exception):
    pass

class NodeLink(QGraphicsPathItem):
    def __init__(self, linktype, startIO=None, endIO=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.linktype = linktype

        self.valid = False

        self.startIO = startIO
        self.endIO = endIO

        if startIO == endIO:
            raise InvalidLinkException("startIO (ID: %s) and endIO (ID: %s) can not be the same!" % (startIO.id, endIO.id))

        if startIO is not None and endIO is not None:
            if issubclass(startIO.iotype, endIO.iotype):
                if type(self.startIO).classMultiplicity == NodeIOMultiplicity.single:
                    for link in self.startIO.nodeLinks:
                        self.startIO.removeLink(link)
                self.startIO.nodeLinks.append(self)

                if type(self.endIO).classMultiplicity == NodeIOMultiplicity.single:
                    for link in self.endIO.nodeLinks:
                        self.endIO.removeLink(link)
                self.endIO.nodeLinks.append(self)

                self.valid = True


    def updateBezier(self, overrideStartpos=QPointF(0, 0), overrideEndpos=QPointF(0, 0)):
        path = QPainterPath()

        if self.startIO is not None:
            startpos = self.startIO.scenePos()
        else:
            startpos = overrideStartpos

        if self.endIO is not None:
            endpos = self.endIO.scenePos()
        else:
            endpos = overrideEndpos

        path.moveTo(startpos)

        controlpoint = QPointF(abs((endpos - startpos).x()) * 0.8, 0)
        path.cubicTo(startpos + controlpoint,
                     endpos - controlpoint,
                     endpos)

        self.setPath(path)



from PyQt5.QtWidgets import QGraphicsItem

from PyQt5.QtWidgets import QGraphicsRectItem
from PyQt5.QtCore import QRectF
from PyQt5.QtGui import QTransform

from enum import Enum
import uuid

from effigy.NodeLink import NodeLink, InvalidLinkException

class NodeIODirection(Enum):
    """Establish if IO direction is input, output or any"""
    any = 0
    output = 1
    input = 2

class NodeIOMultiplicity(Enum):
    """Whether the node can be connected to multiple other nodes, or just one"""
    single = 0
    multiple = 1

class NodeIO(QGraphicsItem):
    """General Input and Output class for Node links.

    This class builds the basis for inputs and outputs for nodes. You link nodes together via those inputs and outputs,
    and all IO must be typed. Only compatible IO (that is, IO of same type) can be linked together. Compatibility is
    established with the 'isinstance' method. Each IO has a direction, i.e. input, output or any. Inputs may only be
    conected with inputs (and 'any' of course), and vice versa."""

    # Direction for this class
    classDirection = NodeIODirection.any
    # Multiplicity for this class
    classMultiplicity = NodeIOMultiplicity.multiple

    def __init__(self, iotype, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if type(iotype) is not type:
            raise TypeError("iotype argument is of type '%s', but must be of type 'type'." % type(iotype))
        self.iotype = iotype
        self.iodirection = NodeIO.classDirection
        self.id = uuid.uuid4().int  # Unique ID to identify node across multiple sessions.

        self.addGraphicsItems()

        self.nodeLinks = []
        self.tempNodeLink = None

    # Define a bounding rect for collision detection (needed for interacting with the IO)
    def boundingRect(self):
        return self.backgroundItem.rect()

    def addGraphicsItems(self):
        """Add Items to draw the IO pin. Can be overridden for custom graphics"""
        self.backgroundItem = QGraphicsRectItem(QRectF(-5, -5, 10, 10), self)

    def paint(self, *__args):
        """Must be defined, but as this is just a container item, it doesn't actually paint anything."""
        pass

    def deleteLink(self, link):
        link.startIO.nodeLinks.remove(link)
        link.endIO.nodeLinks.remove(link)
        self.scene().removeItem(link)

    def mousePressEvent(self, QGraphicsSceneMouseEvent):
        if type(self).classMultiplicity == NodeIOMultiplicity.single:
            for link in self.nodeLinks:
                self.deleteLink(link)
        if self.tempNodeLink is not None:
            pass    # TODO: Throw exception? Can this realistically ever happen?
        else:
            if self.classDirection == NodeIODirection.input:
                # On inputs, io is the ending point of bezier curve
                self.tempNodeLink = NodeLink(self.iotype, startIO=None, endIO=self)
            else:
                # Everywhere else, io is start point of bezier curve
                self.tempNodeLink = NodeLink(self.iotype, startIO=self, endIO=None)
            self.scene().addItem(self.tempNodeLink)


    def mouseMoveEvent(self, QGraphicsSceneMouseEvent):
        # update temporary link
        if self.classDirection == NodeIODirection.input:
            self.tempNodeLink.updateBezier(overrideStartpos=QGraphicsSceneMouseEvent.scenePos())
        else:
            self.tempNodeLink.updateBezier(overrideEndpos=QGraphicsSceneMouseEvent.scenePos())

    def mouseReleaseEvent(self, QGraphicsSceneMouseEvent):
        # Delete temporary link
        self.scene().removeItem(self.tempNodeLink)
        self.tempNodeLink = None

        # Try to find any item at end coordinates
        endpos = QGraphicsSceneMouseEvent.scenePos()
        itematendpos = self.scene().itemAt(endpos, QTransform())

        def recursiveIOCheck(item):
            """Recursively check if this item is a child of an IO item"""
            if item is None:  # We reached the topmost item (or there was none to begin with)
                return None
            if isinstance(item, NodeIO):  # Found a NodeIO
                return item
            else:
                return recursiveIOCheck(item.parentItem())     # We must go deeper and deeper!

        # Check if item is nodeIO
        enditem = recursiveIOCheck(itematendpos)
        if enditem is not None:
            try:
                if self.classDirection == NodeIODirection.input:
                    # On inputs, io is the ending point of bezier curve
                    newLink = NodeLink(self.iotype, startIO=enditem, endIO=self)
                else:
                    # Everywhere else, io is start point of bezier curve
                    newLink = NodeLink(self.iotype, startIO=self, endIO=enditem)

                if newLink.valid:
                    newLink.updateBezier()
                    self.scene().addItem(newLink)
            except InvalidLinkException:
                pass



class NodeInput(NodeIO):
    """General Input class for Node links. See NodeIO class for more information."""
    classDirection = NodeIODirection.input
    classMultiplicity = NodeIOMultiplicity.multiple

class NodeOutput(NodeIO):
    """General Output class for Node links. See NodeIO class for more information."""
    classDirection = NodeIODirection.output
    classMultiplicity = NodeIOMultiplicity.single
